# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Visualize Aineko pipelines as a Mermaid graph."""
import os
import webbrowser

import yaml


def render_mermaid_graph(
    config_path: str,
    direction: str = "LR",
    legend: bool = False,
    render_in_browser: bool = False,
) -> None:
    """Builds mermaid graph from an Aineko pipeline config.

    Args:
        config_path: file path to pipeline yaml file
        direction: direction of the graph.
        legend: include a legend in the graph.
        render_in_browser: Whether to render graph in browser. Prints
        graph to stdout otherwise.
    """
    graph = build_mermaid_from_yaml(
        config_path=config_path, direction=direction, legend=legend
    )
    if render_in_browser is True:
        render_graph_in_browser(graph)

    else:
        print(graph)


def build_mermaid_from_yaml(
    config_path: str, direction: str = "LR", legend: bool = False
) -> str:
    """Builds mermaid graph from an Aineko pipeline config.

    Args:
        config_path: file path to pipeline yaml file
        direction: direction of the graph.
        legend: include a legend in the graph.

    Returns:
        A mermaid graph as a string.
    """
    with open(config_path, "r", encoding="utf-8") as f:
        text = f.read()
    root = yaml.safe_load(text)
    pipeline_name = next(iter(root))
    nodes = root[pipeline_name]["nodes"]

    transitions = []
    for node_name, node_subscriptions in nodes.items():
        if "inputs" in node_subscriptions:
            for input_dataset in node_subscriptions["inputs"]:
                transitions.append({"input": input_dataset, "node": node_name})
        if "outputs" in node_subscriptions:
            for output_dataset in node_subscriptions["outputs"]:
                transitions.append(
                    {"node": node_name, "output": output_dataset}
                )
        if (
            "inputs" not in node_subscriptions.keys()
            and "outputs" not in node_subscriptions.keys()
        ):
            transitions.append({"node": node_name})

    mermaid_transitions = []

    for transition in transitions:
        if (
            not "input" in transition.keys()
            and not "output" in transition.keys()
        ):
            mermaid_transitions.append(
                f"N_{transition['node']}(({transition['node']})):::nodeClass"
            )
        else:
            if "input" in transition.keys():
                src_node = (
                    f"T_{transition['input']}[{transition['input']}]"
                    ":::datasetClass"
                )
                tgt_node = f"N_{transition['node']}(({transition['node']})):::nodeClass"  # pylint: disable=line-too-long
            elif "output" in transition.keys():
                src_node = f"N_{transition['node']}(({transition['node']})):::nodeClass"  # pylint: disable=line-too-long
                tgt_node = (
                    f"T_{transition['output']}[{transition['output']}]"
                    ":::datasetClass"
                )

            mermaid_transitions.append(f"{src_node} -->  {tgt_node}")

    header = f"flowchart {direction}\nclassDef datasetClass "
    header += "fill:#87CEEB\nclassDef nodeClass fill:#eba487"

    if legend:
        footer = (
            "subgraph Legend\nnode((Node)):::nodeClass\ndataset[Dataset]"
            ":::datasetClass\nend\n"
        )
    else:
        footer = ""

    mermaid = "\n".join([header, *mermaid_transitions, footer])
    return mermaid


def render_graph_in_browser(mermaid_graph: str) -> None:
    """Renders a mermaid graph in the default browser.

    Note:
        requires internet connection to load mermaid js.

    Args:
        mermaid_graph: the mermaid graph to render.
    """
    mermaid_html = f"""
    <!DOCTYPE html>
    <html lang="en">
    <body>
        <pre class="mermaid">
        {mermaid_graph}
        </pre>
        <script type="module">
        import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@10/dist/mermaid.esm.min.mjs';
        </script>
    </body>
    </html>
    """
    cwd = os.getcwd()
    with open(f"{cwd}/mermaid.html", "w", encoding="utf-8") as f:
        f.write(mermaid_html)
    webbrowser.open(f"file://{cwd}/mermaid.html")



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Package information for the Aineko package."""


__version__ = "0.2.1"
__author__ = "Convex Labs Engineering"

from aineko.core.config_loader import ConfigLoader
from aineko.core.dataset import (
    DatasetConsumer,
    DatasetProducer,
    FakeDatasetConsumer,
    FakeDatasetProducer,
)
from aineko.core.node import AbstractNode
from aineko.core.runner import Runner



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Aineko command line interface."""
import argparse
import logging

from aineko import __version__
from aineko.cli.create_pipeline import create_pipeline_directory
from aineko.cli.docker_cli_wrapper import DockerCLIWrapper
from aineko.cli.kafka_cli_wrapper import KafkaCLIWrapper
from aineko.cli.run import main as run_main
from aineko.cli.visualize import render_mermaid_graph


def _setup_logging() -> None:
    """Setup logging for the application."""
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s.%(msecs)d - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def _create_parser(subparser: argparse._SubParsersAction) -> None:
    """Subparser for the `aineko create` command.

    Args:
        subparser: Subparser for the `aineko` command.
    """
    create_parser = subparser.add_parser(
        "create",
        help="Create a directory containing pipeline config and \
            code from a template to get started",
    )
    create_parser.add_argument(
        "-d",
        "--deployment-config",
        action="store_true",
        help="Include deploy.yml that facilitates deployment of pipelines",
    )
    create_parser.set_defaults(
        func=lambda args: create_pipeline_directory(args.deployment_config)
    )


def _run_parser(subparser: argparse._SubParsersAction) -> None:
    """Subparser for the `aineko run` command.

    Args:
        subparser: Subparser for the `aineko` command.
    """
    run_parser = subparser.add_parser("run", help="Run a pipeline")
    run_parser.add_argument(
        "config_path",
        help="Path to the config file containing the pipeline config.",
        type=str,
    )
    run_parser.add_argument(
        "-p", "--pipeline", help="Name of the pipeline to run (optional)."
    )
    run_parser.add_argument(
        "-r",
        "--retry",
        help="Retry running the pipeline on failure.",
        action=argparse.BooleanOptionalAction,
        default=False,
    )
    run_parser.set_defaults(
        func=lambda args: run_main(
            pipeline_config_file=args.config_path,
            pipeline_name=args.pipeline,
            retry=args.retry,
        )
    )


def _service_parser(parser: argparse._SubParsersAction) -> None:
    """Subparser for the `aineko service` command."""
    service_parser = parser.add_parser(
        "service", help="Manage Aineko related services"
    )
    service_parser.add_argument(
        "-c",
        "--config",
        help=(
            "Path to the custom Docker Compose config file. Please note that "
            "this is a feature for power users and should be used with caution."
        ),
        type=str,
    )

    service_parser.set_defaults(func=lambda args: service_parser.print_help())

    service_subparsers = service_parser.add_subparsers()

    start_service_parser = service_subparsers.add_parser(
        "start", help="Start Aineko related services"
    )

    start_service_parser.set_defaults(
        func=lambda args: DockerCLIWrapper(args.config).start_service()
    )

    stop_service_parser = service_subparsers.add_parser(
        "stop", help="Stop Aineko related services"
    )

    stop_service_parser.set_defaults(
        func=lambda args: DockerCLIWrapper(args.config).stop_service()
    )

    restart_service_parser = service_subparsers.add_parser(
        "restart", help="Restart Aineko related services"
    )

    restart_service_parser.set_defaults(
        func=lambda args: DockerCLIWrapper(args.config).restart_service()
    )


def _stream_parser(subparser: argparse._SubParsersAction) -> None:
    """Subparser for the `aineko stream` command.

    Args:
        subparser: Subparser for the `aineko` command.
    """
    stream_parser = subparser.add_parser(
        "stream", help="Stream messages from a dataset"
    )

    stream_parser.add_argument(
        "-d", "--dataset", help="Name of dataset", required=True
    )
    stream_parser.add_argument(
        "--from-start",
        help="If we should stream messages from the start or not",
        action="store_true",
    )

    # consume from beginning
    stream_parser.set_defaults(
        func=lambda args: KafkaCLIWrapper.consume_kafka_topic(
            args.dataset, from_beginning=args.from_start
        )
    )


def _visualize_parser(subparser: argparse._SubParsersAction) -> None:
    """Subparser for the `aineko visualize` command.

    Args:
        subparser: Subparser for the `aineko` command.
    """
    visualize_parser = subparser.add_parser(
        "visualize", help="Visualize Aineko pipelines as a Mermaid graph."
    )
    visualize_parser.add_argument(
        "config_path",
        type=str,
        help="Path to the config file containing the pipeline config.",
    )
    visualize_parser.add_argument(
        "-d",
        "--direction",
        type=str,
        default="LR",
        help=(
            "Direction of the graph. Either LR (left to right) or"
            " TD (top down)."
        ),
        choices=["TD", "LR"],
    )
    visualize_parser.add_argument(
        "-l",
        "--legend",
        action="store_true",
        help="Include a legend in the graph.",
    )
    visualize_parser.add_argument(
        "-b",
        "--browser",
        action="store_true",
        help="Open the graph in the default browser.",
    )
    visualize_parser.set_defaults(
        func=lambda args: render_mermaid_graph(
            config_path=args.config_path,
            direction=args.direction,
            legend=args.legend,
            render_in_browser=args.browser,
        )
    )


def _cli() -> None:
    """Command line interface for Aineko."""
    parser = argparse.ArgumentParser(
        prog="aineko",
        description=(
            "Aineko is a framework for building data intensive applications."
        ),
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    subparsers = parser.add_subparsers()
    _create_parser(subparsers)
    _run_parser(subparsers)
    _service_parser(subparsers)
    _stream_parser(subparsers)
    _visualize_parser(subparsers)

    args = parser.parse_args()
    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    _setup_logging()
    logger = logging.getLogger("aineko")
    logger.info("Application is starting.")
    _cli()



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Loads deployment config for an aineko project.

Users define their deployment configuration in a user-friendly format
that is then converted into a full deployment config. The user-friendly
format is more compact, and follows the schema that can be found in
the DeploymentConfig model.

The full deployment config is a comprehensive mapping between
every pipeline and its deployment configuration. It is the source
of truth in which infrastructure should be deployed from.
"""

from collections import defaultdict
from typing import Optional

from pydantic.utils import deep_update

from aineko.models.deploy_config_schema import (
    DeploymentConfig,
    FullDeploymentConfig,
)
from aineko.utils.io import load_yaml


def generate_deploy_config_from_file(
    user_config_file: str, config_type: str = "full"
) -> dict:
    """Generates full or user deployment config from deployment config file.

    Args:
        user_config_file: path to the deployment config file
        type: `full` or `user`. If `full`, returns the full deployment config

    Returns:
        Full or user deployment configuration.
    """
    user_config = load_yaml(user_config_file)
    config = generate_deploy_config(user_config, config_type=config_type)
    return config


def generate_deploy_config(
    user_config: dict, config_type: str = "full"
) -> dict:
    """Generates full or user deployment config from the input user config dict.

    Args:
        user_config: deployment configuration in the user-friendly
            format. See DeploymentConfig for expected schema.

    Returns:
        Full or user deployment configuration.
    """
    if config_type not in ["full", "user"]:
        raise ValueError(
            "Specified output config type must be either `full` or `user`"
        )

    user_deploy_config = DeploymentConfig(**user_config)
    if config_type == "user":
        return user_deploy_config.dict()

    else:
        full_config = _generate_full_config(user_deploy_config)
        return full_config.dict()


def _generate_full_config(
    user_config: Optional[DeploymentConfig] = None,
) -> FullDeploymentConfig:
    """Generates a full deployment config from the user config.

    For each pipeline specified in each environment, we start with the
    default config, then override with all pipeline-specific config,
    then override with all environment-specific config.

    Args:
        user_config: deployment configuration in the user-friendly format.
        See DeploymentConfig for expected schema.

    Returns:
        Full deployment configuration.
    """
    user_config = user_config or user_config
    if not user_config:
        raise ValueError("User config has not been loaded nor defined.")
    full_config: dict = {
        "version": user_config.version,
        "environments": defaultdict(lambda: {"pipelines": []}),
    }
    for env, env_pipelines in user_config.environments.items():
        for pipeline in env_pipelines.pipelines:
            if isinstance(pipeline, str):
                pipeline_name = pipeline
            else:
                # Get first key in dict
                pipeline_name = next(iter(pipeline))

            defaults = (
                user_config.defaults.dict() if user_config.defaults else {}
            )
            pipeline_specific_config = {
                k: v
                for k, v in user_config.pipelines[pipeline_name].dict().items()
                if v
            }

            # Env specific config may not be defined
            if isinstance(pipeline, dict):
                env_specific_config = {
                    k: v for k, v in pipeline[pipeline_name].dict().items() if v
                }
            else:
                env_specific_config = {}

            # Env-specific overwrites pipeline-specific overwrites defaults
            full_config["environments"][env]["pipelines"].append(
                {
                    pipeline_name: deep_update(
                        defaults,
                        pipeline_specific_config,
                        env_specific_config,
                    )
                }
            )

        # Add any defined load balancers
        full_config["environments"][env][
            "load_balancers"
        ] = env_pipelines.load_balancers
    return FullDeploymentConfig(**full_config)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Module to load config files."""
import logging
from typing import Union, overload

from pydantic import ValidationError

from aineko.config import AINEKO_CONFIG
from aineko.models.config_schema import Config
from aineko.utils.io import load_yaml

logger = logging.getLogger(__name__)


class ConfigLoader:
    """Class to read yaml config files.

    Args:
        pipeline_config_file: path of pipeline config file. Defaults
        to `DEFAULT_PIPELINE_CONFIG`.

    Attributes:
        pipeline_config_file (str): path to the pipeline configuration file
        config_schema (Config): Pydantic model to validate a pipeline config

    Methods:
        load_config: loads and validates the pipeline config from a yaml file
    """

    def __init__(
        self,
        pipeline_config_file: str,
    ):
        """Initialize ConfigLoader."""
        self.pipeline_config_file = pipeline_config_file or AINEKO_CONFIG.get(
            "DEFAULT_PIPELINE_CONFIG"
        )

        # Setup config schema
        self.config_schema = Config

    def load_config(self) -> dict:
        """Load and validate the pipeline config.

        Raises:
            ValidationError: If the config does not match the schema

        Returns:
            The pipeline config as a dictionary
        """
        config = load_yaml(self.pipeline_config_file)

        try:
            Config(**config)
        except ValidationError as e:
            logger.error(
                "Schema validation failed for pipeline `%s` loaded from %s. "
                "See detailed error below.",
                config["pipeline"]["name"],
                self.pipeline_config_file,
            )
            raise e

        return config

    @overload
    def _update_params(self, value: dict, params: dict) -> dict:
        ...

    @overload
    def _update_params(self, value: list, params: dict) -> list:
        ...

    @overload
    def _update_params(self, value: str, params: dict) -> str:
        ...

    @overload
    def _update_params(self, value: int, params: dict) -> int:
        ...

    def _update_params(
        self, value: Union[dict, list, str, int], params: dict
    ) -> Union[dict, list, str, int]:
        """Update value with params.

        Recursively calls the method if value is a list or dictionary until it
        reaches a string or int. If string then formats the str with variable
        mapping in params dict.

        Args:
            value: value to update
            params: params to update value with

        Returns:
            object with updated values (dict, list, str, or int)
        """
        if isinstance(value, dict):
            new_dict_val = {}
            for key, val in value.items():
                new_dict_val[key] = self._update_params(val, params)
            return new_dict_val
        if isinstance(value, list):
            new_list_val: list = []
            for val in value:
                new_list_val.append(self._update_params(val, params))
            return new_list_val
        if isinstance(value, str):
            for key, val in params.items():
                value = value.replace(f"${key}", val)
            return value
        if isinstance(value, (int, float)):
            return value
        raise ValueError(
            f"Invalid value type {type(value)}. "
            "Expected dict, list, str, or int."
        )



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Utils to create a aineko pipeline from existing templates."""

import os

from cookiecutter.main import cookiecutter  # type: ignore


def create_pipeline_directory(deployment_config: bool) -> None:
    """Creates boilerplate code and config required for an Aineko pipeline.

    Args:
        deployment_config: If True, include a deploy file when generating, else
        do not include.
    """
    script_directory = os.path.dirname(os.path.abspath(__file__))
    # assumes templates is a folder that is one-level up
    templates_directory = f"{os.path.dirname(script_directory)}/templates"
    # Only one type of pipeline, hence this argument is no-arg
    cookiecutter(
        f"{templates_directory}/first_aineko_pipeline",
        extra_context={
            "_deployment_config": deployment_config,
        },
    )



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Configuration file for Aineko modules.

Kafka configuration can be set using the following environment variables:

KAFKA_CONFIG: JSON string with kafka configuration
(see https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
for all options)

Additionally, the following environment variables can be used to specify certain
configuration values. They correspond to configuration keys found in the above
link, but with a prefix. For example, `KAFKA_CONFIG_BOOTSTRAP_SERVERS`
corresponds to `bootstrap.servers`.

- KAFKA_CONFIG_BOOTSTRAP_SERVERS (e.g. localhost:9092,localhost:9093)
- KAFKA_CONFIG_SASL_USERNAME
- KAFKA_CONFIG_SASL_PASSWORD
- KAFKA_CONFIG_SECURITY_PROTOCOL
- KAFKA_CONFIG_SASL_MECHANISM
"""
import copy
import json
import os
from typing import Any, Dict


# pylint: disable=too-few-public-methods
# pylint: disable=invalid-name
class BaseConfig:
    """Base Config."""

    @classmethod
    def get(cls, attribute: str) -> Any:
        """Returns the base config."""
        if hasattr(cls, attribute):
            return copy.deepcopy(getattr(cls, attribute))
        raise ValueError(f"{attribute} not found in {cls.__name__}")


class DEFAULT_KAFKA_CONFIG(BaseConfig):
    """Kafka configuration."""

    # Default Kafka broker settings
    BROKER_CONFIG = {
        "bootstrap.servers": "localhost:9092",
    }

    kafka_config = os.environ.get("KAFKA_CONFIG", "{}")
    BROKER_CONFIG.update(json.loads(kafka_config))

    # Override these fields if set
    OVERRIDABLES = {
        "KAFKA_CONFIG_BOOTSTRAP_SERVERS": "bootstrap.servers",
        "KAFKA_CONFIG_SASL_USERNAME": "sasl.username",
        "KAFKA_CONFIG_SASL_PASSWORD": "sasl.password",
        "KAFKA_CONFIG_SECURITY_PROTOCOL": "security.protocol",
        "KAFKA_CONFIG_SASL_MECHANISM": "sasl.mechanism",
    }
    for env, config in OVERRIDABLES.items():
        value = os.environ.get(env)
        if value:
            BROKER_CONFIG[config] = value

    # Config for default kafka consumer
    CONSUMER_CONFIG: Dict[str, str] = {
        **BROKER_CONFIG,
        "auto.offset.reset": "earliest",
    }

    # Config for default kafka producer
    PRODUCER_CONFIG: Dict[str, str] = {**BROKER_CONFIG}

    # Default dataset config
    DATASET_PARAMS = {
        # One single partition for each dataset
        "num_partitions": 1,
        # No replication
        "replication_factor": 1,
        "config": {
            # Keep messages for 7 days
            "retention.ms": 1000
            * 60
            * 60
            * 24
            * 7,
        },
    }

    # Default Kafka consumer settings
    # Timeout for kafka consumer polling (seconds)
    CONSUMER_TIMEOUT = 0
    # Max number of messages to retreive when getting the last message
    CONSUMER_MAX_MESSAGES = 1000000

    # Default Kafka producer settings
    # Producer overridables
    # See: https://kafka.apache.org/documentation/#producerconfigs
    # Empty list means no overridable settings
    PRODUCER_OVERRIDABLES = []  # type: ignore

    # Default datasets to create for every pipeline
    LOGGING_DATASET = "logging"
    DATASETS = [LOGGING_DATASET]


class TESTING_NODE_CONFIG(BaseConfig):
    """Testing node configuration."""

    DATASETS = DEFAULT_KAFKA_CONFIG.get("DATASETS")


class AINEKO_CONFIG(BaseConfig):
    """Aineko configuration."""

    # Port to expose metrics
    RAY_METRICS_PORT = 8080

    # Aineko monitoring config
    HEARTBEAT_INTERVAL = 1

    # Timeout in seconds for dataset creation
    DATASET_CREATION_TIMEOUT = 300

    DEFAULT_PIPELINE_CONFIG = os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "conf/pipeline.yml"
        )
    )
    MSG_TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
    KAFKA_STREAM_TYPE = "kafka_stream"

    # Default cpu for each node
    DEFAULT_NUM_CPUS = 0.5

    # Valid log levels
    LOG_LEVELS = ("info", "debug", "warning", "error", "critical")


class NODE_MANAGER_CONFIG(BaseConfig):
    """Node Manager configuration."""

    # Name to call node
    NAME = "node_manager"

    # Ray options for NodeManager
    RAY_OPTIONS = {
        "num_cpus": 0.1,
    }

    # Node config
    NODE_CONFIG = {
        "class": "aineko.core.node_manager.NodeManager",
    }



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A series of cookiecutter post-codegen hooks to validate user input.

These code will run after files are generated

See: https://cookiecutter.readthedocs.io/en/1.7.2/advanced/hooks.html
"""
import os
import sys

# Color codes for logging
ERROR = "\x1b[1;31m[ERROR]: "
TERMINATOR = "\x1b[0m"


project_slug = "{{cookiecutter.project_slug}}"
with_deploy_str = "{{ cookiecutter._deployment_config }}"

# Cast string to boolean
if with_deploy_str.lower() == "true":
    with_deploy = True
elif with_deploy_str.lower() == "false":
    with_deploy = False
else:
    with_deploy = False
    print(
        ERROR + f"Invalid value for cookiecutter._deployment_config:"
        f" {with_deploy_str}" + TERMINATOR
    )
    sys.exit(1)


def validate_deploy_file_path(file_path: str) -> None:
    """Validates that path to deploy.yaml points to a file.

    Args:
        file_path: The path to deploy.yml file

    Raises:
        FileNotFoundError: If the 'deploy.yml' file does not exist.
    """
    # Check if the file exists
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"The file '{file_path}' does not exist.")


def _get_deploy_file_path() -> str:
    # The hook runs in the newly created project
    file_path = os.path.join(os.getcwd(), "deploy.yml")
    return file_path


try:
    deploy_file_path = _get_deploy_file_path()
    validate_deploy_file_path(deploy_file_path)
    if not with_deploy:
        os.remove(deploy_file_path)


except FileNotFoundError as ex:
    print(ERROR + str(ex) + TERMINATOR)
    sys.exit(1)



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Node base class and poison pill global actor.

Global variables that can be accessed by multiple nodes
should be stored in an actor instance (see
https://stackoverflow.com/questions/67457237/share-object-between-actors-in-ray).
The PoisonPill actor stores the boolean value that represents
if it should be activated or not. Upon activation, the Node Manager
will kill the entire pipeline. All nodes have access to this variable,
and can activate it by calling the `activate_poison_pill` method.

The AbstractNode class is the parent class for all Aineko pipeline nodes.
It contains  helper methods for setup, util methods, and dummy methods
that users should override with their own implementation.
"""
import time
import traceback
from abc import ABC, abstractmethod
from typing import Dict, Generator, List, Optional, Tuple

import ray

from aineko.config import (
    AINEKO_CONFIG,
    DEFAULT_KAFKA_CONFIG,
    TESTING_NODE_CONFIG,
)
from aineko.core.dataset import (
    DatasetConsumer,
    DatasetProducer,
    FakeDatasetConsumer,
    FakeDatasetProducer,
)


class PoisonPill:
    """Global variable accessible to every node in pipeline.

    This is the recommended approach to share objects between
    Ray Actors.
    """

    def __init__(self) -> None:
        """Poison pill initializes with unactivated state."""
        self.state = False

    def activate(self) -> None:
        """Activate poison pill by setting state as True."""
        self.state = True

    def get_state(self) -> bool:
        """Gets poison pill state."""
        return self.state


class AbstractNode(ABC):
    """Node base class for all nodes in the pipeline.

    Nodes are the building blocks of the pipeline and are responsible for
    executing the pipeline. Nodes are designed to be modular and can be
    combined to create a pipeline. The node base class provides helper methods
    for setting up the consumer and producer for a node. The execute method is
    a wrapper for the _execute method which is to be implemented by subclasses.
    The _execute method is where the node logic is implemented by the user.

    Attributes:
        consumers (dict): dict of DatasetConsumer objects for inputs to node
        producers (dict): dict of DatasetProducer objects for outputs of node
        last_hearbeat (float): timestamp of the last heartbeat
        test (bool): True if node is in test mode else False
        log_levels (tuple): tuple of allowed log levels
        local_state (dict): shared local state between nodes. Used for intra-
            pipeline communication without dataset dependency.

    Methods:
        setup_datasets: setup the consumers and producers for a node
        execute: execute the node, wrapper for _execute method
        _execute: execute the node, to be implemented by subclasses
    """

    def __init__(
        self,
        poison_pill: Optional[ray.actor.ActorHandle] = None,
        test: bool = False,
    ) -> None:
        """Initialize the node."""
        self.last_heartbeat = time.time()
        self.consumers: Dict = {}
        self.producers: Dict = {}
        self.params: Dict = {}
        self.test = test
        self.log_levels = AINEKO_CONFIG.get("LOG_LEVELS")
        self.poison_pill = poison_pill

    def enable_test_mode(self) -> None:
        """Enable test mode."""
        self.test = True

    def setup_datasets(
        self,
        datasets: Dict[str, dict],
        node: str,
        pipeline: str,
        inputs: Optional[List[str]] = None,
        outputs: Optional[List[str]] = None,
        prefix: Optional[str] = None,
        has_pipeline_prefix: bool = False,
    ) -> None:
        """Setup the consumer and producer for a node.

        Args:
            datasets: dataset configuration
            node: name of the node
            pipeline: name of the pipeline
            inputs: list of dataset names for the inputs to the node
            outputs: list of dataset names for the outputs of the node
            prefix: prefix for topic name (`<prefix>.<dataset_name>`)
            has_pipeline_prefix: whether the dataset name has pipeline name
            prefix
        """
        inputs = inputs or []
        self.consumers.update(
            {
                dataset_name: DatasetConsumer(
                    dataset_name=dataset_name,
                    node_name=node,
                    pipeline_name=pipeline,
                    dataset_config=datasets.get(dataset_name, {}),
                    prefix=prefix,
                    has_pipeline_prefix=has_pipeline_prefix,
                )
                for dataset_name in inputs
            }
        )

        outputs = outputs or []
        self.producers.update(
            {
                dataset_name: DatasetProducer(
                    dataset_name=dataset_name,
                    node_name=node,
                    pipeline_name=pipeline,
                    dataset_config=datasets.get(dataset_name, {}),
                    prefix=prefix,
                    has_pipeline_prefix=has_pipeline_prefix,
                )
                for dataset_name in outputs
            }
        )

    def setup_test(
        self,
        inputs: Optional[dict] = None,
        outputs: Optional[list] = None,
        params: Optional[dict] = None,
    ) -> None:
        """Setup the node for testing.

        Args:
            inputs: inputs to the node, format should be {"dataset": [1, 2, 3]}
            outputs: outputs of the node, format should be
                ["dataset_1", "dataset_2", ...]
            params: dictionary of parameters to make accessible to _execute

        Raises:
            RuntimeError: if node is not in test mode
        """
        if self.test is False:
            raise RuntimeError(
                "Node is not in test mode. "
                "Please initialize with `enable_test_mode()`."
            )

        inputs = inputs or {}
        self.consumers = {
            dataset_name: FakeDatasetConsumer(
                dataset_name=dataset_name,
                node_name=self.__class__.__name__,
                values=values,
            )
            for dataset_name, values in inputs.items()
        }
        outputs = outputs or []
        outputs.extend(TESTING_NODE_CONFIG.get("DATASETS"))
        self.producers = {
            dataset_name: FakeDatasetProducer(
                dataset_name=dataset_name,
                node_name=self.__class__.__name__,
            )
            for dataset_name in outputs
        }
        self.params = params or {}

    def log(self, message: str, level: str = "info") -> None:
        """Log a message to the logging dataset.

        Args:
            message: Message to log
            level: Logging level. Defaults to "info". Options are:
                "info", "debug", "warning", "error", "critical"
        Raises:
            ValueError: if invalid logging level is provided
        """
        if level not in self.log_levels:
            raise ValueError(
                f"Invalid logging level {level}. Valid options are: "
                f"{', '.join(self.log_levels)}"
            )
        out_msg = {"log": message, "level": level}
        self.producers[DEFAULT_KAFKA_CONFIG.get("LOGGING_DATASET")].produce(
            out_msg
        )

    def catch_exception(self) -> None:
        """Catch an exception and report it.

        Args:
            exc: Exception to catch
        """
        exc_info = traceback.format_exc()
        self.log(exc_info, level="error")

    def execute(self, params: Optional[dict] = None) -> None:
        """Execute the node.

        Wrapper for _execute method to be implemented by subclasses.

        Args:
            params: Parameters to use to execute the node. Defaults to None.
        """
        params = params or {}
        run_loop = True

        try:
            self._pre_loop_hook(params)
        except Exception:  # pylint: disable=broad-except
            self.catch_exception()
            raise

        while run_loop is not False:
            # Monitoring
            try:
                run_loop = self._execute(params)  # type: ignore
            except Exception:  # pylint: disable=broad-except
                self.catch_exception()
                raise

        self.log(f"Execution loop complete for node: {self.__class__.__name__}")
        self._post_loop_hook(params)

    def activate_poison_pill(self) -> None:
        """Activates poison pill, shutting down entire pipeline."""
        if self.poison_pill:
            ray.get(self.poison_pill.activate.remote())

    @abstractmethod
    def _execute(self, params: dict) -> Optional[bool]:
        """Execute the node.

        Args:
            params: Parameters to use to execute the node.

        Note:
            Method to be implemented by subclasses

        Raises:
            NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError("_execute method not implemented")

    def run_test(self, runtime: Optional[int] = None) -> dict:
        """Execute the node in testing mode.

        Runs the steps in execute that involves the user defined methods.
        Includes pre_loop_hook, _execute, and post_loop_hook.

        Args:
            runtime: Number of seconds to run the execute loop for.

        Returns:
            dict: dataset names and values produced by the node.
        """
        if self.test is False:
            raise RuntimeError(
                "Node is not in test mode. "
                "Please initialize with `enable_test_mode()`."
            )
        run_loop = True
        start_time = time.time()

        self._pre_loop_hook(self.params)
        while run_loop is not False:
            run_loop = self._execute(self.params)  # type: ignore

            # Do not end loop if runtime not exceeded
            if runtime is not None:
                if time.time() - start_time < runtime:
                    continue

            # End loop if all consumers are empty
            if self.consumers and all(
                consumer.empty for consumer in self.consumers.values()
            ):
                run_loop = False

        self._post_loop_hook(self.params)

        return {
            dataset_name: producer.values
            for dataset_name, producer in self.producers.items()
        }

    def run_test_yield(
        self, runtime: Optional[int] = None
    ) -> Generator[Tuple[dict, dict, "AbstractNode"], None, None]:
        """Execute the node in testing mode, yielding at each iteration.

        This method is an alternative to `run_test`. Instead of returning the
        aggregated output, it yields the most recently consumed value, the
        produced value and the current node instance at each iteration. This is
        useful for testing nodes that either don't produce any output or if you
        need to test intermediate outputs. Testing state modifications is also
        possible using this method.

        Args:
            runtime: Number of seconds to run the execute loop for.

        Yields:
            A tuple containing the most recent input value, output value and
            the node instance.

        Example:
            >>> for input, output, node_instance in sequencer.run_test_yield():
            >>>     print(f"Input: {input}, Output: {output})
            >>>     print(f"Node Instance: {node_instance}")
        """
        if self.test is False:
            raise RuntimeError(
                "Node is not in test mode. "
                "Please initialize with `enable_test_mode()`."
            )
        run_loop = True
        start_time = time.time()

        self._pre_loop_hook(self.params)
        while run_loop is not False:
            last_produced_values = {}
            last_consumed_values = {}

            run_loop = self._execute(self.params)  # type: ignore

            # Do not end loop if runtime not exceeded
            if runtime is not None:
                if time.time() - start_time < runtime:
                    continue

            # End loop if all consumers are empty
            if self.consumers and all(
                consumer.empty for consumer in self.consumers.values()
            ):
                run_loop = False

            # Capture last consumed values
            for dataset_name, consumer in self.consumers.items():
                if consumer.values:
                    last_value = consumer.values[0]
                    last_consumed_values[dataset_name] = last_value

            # Capture last produced values
            for dataset_name, producer in self.producers.items():
                if producer.values:
                    last_value = producer.values[-1]
                    last_produced_values[dataset_name] = last_value

            yield (last_consumed_values, last_produced_values, self)

        self._post_loop_hook(self.params)

    def _pre_loop_hook(self, params: Optional[dict] = None) -> None:
        """Hook to be called before the node loop. User overrideable.

        Args:
            params: Parameters to use to execute the node.

        Note:
            Method (optional) to be implemented by subclasses.
        """
        pass

    def _post_loop_hook(self, params: Optional[dict] = None) -> None:
        """Hook to be called after the node loop. User overrideable.

        Args:
            params: Parameters to use to execute the node.

        Note:
            Method (optional) to be implemented by subclasses.
        """
        pass



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Node code for Node Manager.

The Node Manager is a special node that exists in every
pipeline and is responsible for the following tasks:
 - Killing the ray instance when the poison pill is activated
"""

import time
from typing import Optional

import ray

from aineko.core.node import AbstractNode


class NodeManager(AbstractNode):
    """Manages all other nodes in pipeline."""

    def _execute(self, params: Optional[dict] = None) -> None:
        """Kills all other nodes."""
        # Sleep to prevent high CPU usage
        time.sleep(0.1)
        if self.poison_pill:
            if ray.get(self.poison_pill.get_state.remote()):
                self.log("Poison pill activated, killing all nodes.")
                ray.shutdown()



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""IO utilities."""
from typing import Union

import yaml


def load_yaml(file_path: str) -> dict:
    """Load yaml file into a dictionary.

    Args:
        file_path: path to yaml file

    Returns:
        dictionary of yaml file
    """
    with open(file_path, "r", encoding="utf-8") as conf_file:
        contents = yaml.safe_load(conf_file)
    if not contents:
        raise ValueError(f"YAML file {file_path} cannot be loaded.")
    return contents


def load_yamls(file_paths: Union[str, list]) -> dict:
    """Load yaml file(s) into a dictionary.

    If multiple keys with the same name exist, the last one will be used.

    Args:
        file_paths: list of paths to yaml files

    Returns:
        dictionary of combined yaml files
    """
    if isinstance(file_paths, str):
        file_paths = [file_paths]

    yaml_dict = {}
    for file_path in file_paths:
        yaml_dict.update(load_yaml(file_path))
    return yaml_dict



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""Internal models for deployment configuration."""

from typing import Dict, List, Optional, Union

from pydantic import BaseModel, validator

from aineko.models.validations import check_power_of_2


class MachineConfig(BaseModel, extra="forbid"):
    """Configuration for cloud machine that runs pipelines."""

    type: str
    mem_gib: int
    vcpu: int

    @validator("mem_gib")
    def memory(cls, value: int) -> int:  # pylint: disable=no-self-argument
        """Validates that memory is a power of 2."""
        return check_power_of_2(value)

    @validator("vcpu")
    def power_of_2(cls, value: int) -> int:  # pylint: disable=no-self-argument
        """Validates that vcpu is a power of 2."""
        return check_power_of_2(value)


class ParameterizableDefaults(BaseModel, extra="forbid"):
    """Parameters that can be set in the defaults block."""

    machine_config: Optional[MachineConfig]
    env_vars: Optional[Dict[str, str]]


class GenericPipeline(BaseModel, extra="forbid"):
    """Configuration for a pipeline defined under top-level pipelines key."""

    source: str
    name: Optional[str]
    machine_config: Optional[MachineConfig]
    env_vars: Optional[Dict[str, str]]


class LoadBalancer(BaseModel, extra="forbid"):
    """Configuration for a load balancer."""

    pipeline: str
    port: int


class SpecificPipeline(BaseModel, extra="forbid"):
    """Pipeline defined under the top-level environments key."""

    source: Optional[str]  # Pipeline config file path
    name: Optional[str]  # Pipeline name
    machine_config: Optional[MachineConfig]
    env_vars: Optional[Dict[str, str]]
    load_balancers: Optional[List[LoadBalancer]]


class FullPipeline(BaseModel, extra="forbid"):
    """Pipeline defined in the full deployment config."""

    source: str
    name: Optional[str]
    machine_config: MachineConfig
    env_vars: Optional[Dict[str, str]]


class Environment(BaseModel, extra="forbid"):
    """Environment defined under the top-level environments key."""

    pipelines: List[Union[str, Dict[str, SpecificPipeline]]]
    load_balancers: Optional[Dict[str, List[LoadBalancer]]]


class FullEnvironment(BaseModel, extra="forbid"):
    """Environment defined under the top-level environments key."""

    pipelines: List[Union[Dict[str, FullPipeline], str]]
    load_balancers: Optional[Dict[str, List[LoadBalancer]]]



# Copyright 2023 Aineko Authors
# SPDX-License-Identifier: Apache-2.0
"""A wrapper class that executes Docker CLI commands via subprocess."""
import subprocess
import sys
from typing import Optional


class DockerCLIWrapper:
    """A wrapper class that executes Docker CLI commands via subprocess.

    This class provides methods to start, stop, and restart Docker services
    using docker-compose.

    Methods:
        start_service(cls) -> None:
            Start the Docker service.

        stop_service(cls) -> None:arg
            Stop the running Docker service.

        restart_service(cls) -> None:
            Restart the running Docker service.
    """

    _docker_compose_config = """
version: "3"
services:
  zookeeper:
    image: confluentinc/cp-zookeeper:7.3.0
    hostname: zookeeper
    container_name: zookeeper
    environment:
      ZOOKEEPER_CLIENT_PORT: 2181
      ZOOKEEPER_TICK_TIME: 2000

  broker:
    image: confluentinc/cp-kafka:7.3.0
    container_name: broker
    ports:
      - "9092:9092"
    depends_on:
      - zookeeper
    environment:
      KAFKA_BROKER_ID: 1
      KAFKA_ZOOKEEPER_CONNECT: "zookeeper:2181"
      KAFKA_LISTENER_SECURITY_PROTOCOL_MAP: PLAINTEXT:PLAINTEXT,PLAINTEXT_INTERNAL:PLAINTEXT
      KAFKA_ADVERTISED_LISTENERS: PLAINTEXT://localhost:9092,PLAINTEXT_INTERNAL://broker:29092
      KAFKA_OFFSETS_TOPIC_REPLICATION_FACTOR: 1
      KAFKA_TRANSACTION_STATE_LOG_MIN_ISR: 1
      KAFKA_TRANSACTION_STATE_LOG_REPLICATION_FACTOR: 1
"""

    def __init__(self, custom_config_path: Optional[str] = None) -> None:
        """Initialize DockerCLIWrapper class."""
        if custom_config_path:
            self._load_custom_config(custom_config_path)

    @classmethod
    def start_service(cls) -> None:
        """Start the Docker service."""
        try:
            output = subprocess.check_output(
                args="docker-compose -f - up -d",
                input=cls._docker_compose_config,
                shell=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
            print(output)
        except subprocess.CalledProcessError as ex:
            print(f"Error: {ex}")
            print(f"Command Output: {ex.output}")

    @classmethod
    def stop_service(cls) -> None:
        """Stop the running Docker service."""
        try:
            output = subprocess.check_output(
                args="docker-compose -f - stop",
                input=cls._docker_compose_config,
                shell=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
            print(output)
        except subprocess.CalledProcessError as ex:
            print(f"Error: {ex}")
            print(f"Command Output: {ex.output}")

    @classmethod
    def restart_service(cls) -> None:
        """Restart the running Docker service."""
        try:
            output = subprocess.check_output(
                args="docker-compose -f - restart",
                input=cls._docker_compose_config,
                shell=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
            print(output)
        except subprocess.CalledProcessError as ex:
            print(f"Error: {ex}")
            print(f"Command Output: {ex.output}")

    @classmethod
    def _load_custom_config(cls, custom_config_path: str) -> None:
        """Load a custom Docker Compose config file.

        Args:
            custom_config_path: Path to the custom Docker Compose config file.
        """
        try:
            with open(
                file=custom_config_path, mode="r", encoding="utf-8"
            ) as file:
                cls._docker_compose_config = file.read()
        except FileNotFoundError:
            print(
                f"FileNotFoundError: Custom config file `{custom_config_path}`"
                f" not found."
            )
            sys.exit(1)
